#include "Game.h"

Game::Game(StateMachine* machine) : StateLoopIFC(machine)
{
    artist = Artist::Instance();
    vinylScratch = WubManager::Instance();

    twi = NULL;

    level = NULL;
    camera = new Camera();

    Rect cam;
    cam.x = 0; cam.y = 0;
    cam.w = 100; cam.h = 100;
    camera->setCamera(cam);

    winSound = vinylScratch->requestSoundSource("sounds/win.wav");
    winSound->setDefaultVolume(.3f);
}

Game::~Game()
{
    if(twi)
        delete twi;

    if(level)
        delete level;
}

void Game::spawnPlayer()
{
    if(twi)
        delete twi;

    twi = new Player("sprites/twilight.png");
    twi->ponySetup();

    twi->setPosition(level->getSpawnX(), level->getSpawnY());
    cout << "twi x " << twi->getX() << " y " << twi->getY() << "\n";

    if(stateMachine->getArrowStates(RIGHT))
    {
        twi->faceRight(true);
        twi->setIntendedXAcc(VELOCITY_SCALE/4);
    }
    else if(stateMachine->getArrowStates(LEFT))
    {
        twi->faceRight(false);
        twi->setIntendedXAcc(-VELOCITY_SCALE/4);
    }

    camera->centerOn(twi->getPosition());
    camera->validate(level->getWidth(), level->getHeight());
}

bool Game::newGame(string file)
{
    if(level)
        delete level;

    level = new Level();
    level->loadFile(file);

    Rect cam;
    cam.x = 0;
    cam.y = 0;
    cam.w = level->getBgSize().w;
    cam.h = level->getBgSize().h;
    camera->setCamera(cam);

    spawnPlayer();

    stateMachine->setGameState(CONTINUE);
    return true;
}

void Game::display()
{
    if(level)
        level->displayBg(camera->getBounds(), offsetX, offsetY);

    if(twi)
        twi->display(offsetX - camera->getBounds().x, offsetY - camera->getBounds().y);

    if(level)
        level->displayFg(camera->getBounds(), offsetX, offsetY);
}

void Game::update(int value)
{
    // gravity
    twi->yAcc(VELOCITY_SCALE/2);
    twi->act();
    updateObject(twi);

    if(fatalCollision(twi->getPosition()))
    {
        spawnPlayer();
        twi->playTeleportSound();
    }
    // spawnPlayer takes care of this first
    else
    {
        camera->centerOn(twi->getPosition());
        camera->validate(level->getWidth(), level->getHeight());
        checkLevelFinished();
    }


    glutPostRedisplay();
}

void Game::updateObject(Object* obj)
{
    Rect startPoint = obj->getPosition();
    Rect endPoint = obj->update();

    int xdiff = endPoint.x - startPoint.x;
    int ydiff = endPoint.y - startPoint.y;

    if(!xdiff && !ydiff)
        return; // no change in position

    int sign = -1;

    if( abs(xdiff) > abs(ydiff))
    {
        if(xdiff>0)
            sign = 1;

        obj->setPosition(iteratePosition(startPoint, endPoint, true, (float)ydiff/xdiff, sign));
    }
    else
    {
        if (ydiff>0)
            sign =1;

        obj->setPosition(iteratePosition(startPoint, endPoint, false, (float)xdiff/ydiff, sign));
    }


    if(nonFatalCollision(obj->getXNudge()))
        obj->xBump();

    if(nonFatalCollision(obj->getYNudge()))
    {
        obj->yBump();
        obj->setGrounded(ydiff>=0);
    }
    else
        obj->setGrounded(false);
}

bool Game::fatalCollision(Rect box)
{
    int w, h;
    level->getSize(w, h);

    if( box.y+box.h > h )
        return true;

    const vector<Trap*>* traps = level->getTraps();

    for(unsigned int i = 0; i < traps->size(); i++)
    {
        Rect dangerzone = traps->at(i)->getPosition();

        if(box.x < dangerzone.x + dangerzone.w && box.x + box.w > dangerzone.x && box.y < dangerzone.y + dangerzone.h && box.y + box.h > dangerzone.y)
            return true;

    }

    return false;
}

void Game::checkLevelFinished()
{
    Rect box = twi->getPosition();
    Rect zone = level->getWinZone();

    if(box.x < zone.x + zone.w && box.x + box.w > zone.x && box.y < zone.y + zone.h && box.y + box.h > zone.y)
    {
        vinylScratch->getDJ()->stopAllSounds();
        vinylScratch->getDJ()->play2D(winSound);

        string next = level->getnextLevel();
        if(next == "finish")
        {
            stateMachine->setGameState(GAMEDONE);
            return;
        }

        newGame(next);
    }
}

bool Game::nonFatalCollision(Rect box)
{
    int w, h;
    level->getSize(w, h);

    if(box.y < 0 || box.x < 0 || box.x + box.w > w)
        return true;

    const vector<Platform*>* plats = level->getPlatforms();


    for(unsigned int i = 0; i < plats->size(); i++)
    {
        Rect platform = plats->at(i)->getPosition();

        if(box.x < platform.x + platform.w && box.x + box.w > platform.x && box.y < platform.y + platform.h && box.y + box.h > platform.y)
            return true;
    }

    return false;
}

Rect Game::iteratePosition(const Rect& startPoint, Rect resultPoint, bool xAxis, float slope, int sign)
{
    int startA = startPoint.x;
    int startB = startPoint.y;
    int endA = resultPoint.x;
    int a, b;
    int* x = &a;
    int* y = &b;

    if(!xAxis)
    {
        startA = startPoint.y;
        startB = startPoint.x;
        endA = resultPoint.y;
        x = &b;
        y = &a;
    }

    // keep moving one pixel away from the collision zone
    for(a = endA; a*sign > startA*sign; a-=sign)
    {
        b = startB + (a - startA) * slope;
        resultPoint.x = *x;
        resultPoint.y = *y;
        // stop as soon as it isn't colliding any more
        if(!nonFatalCollision(resultPoint))
            return resultPoint;
    }

    return startPoint;
}

void Game::key(unsigned char key, int x, int y)
{
    switch (key)
    {
        case 27: stateMachine->setGameState(MENU); vinylScratch->getDJ()->stopAllSounds(); break;
        case ' ': if(nonFatalCollision(twi->getYNudge(1)))
                  {
                      twi->jump();
                  } break;
        default: break;
    }
}

void Game::specialDown(int key, int x, int y)
{
    switch(key) {
        case GLUT_KEY_RIGHT : stateMachine->setArrowState(RIGHT, true);
                                        twi->faceRight(true);
                                        twi->setIntendedXAcc(VELOCITY_SCALE/4);
                                        break;
        case GLUT_KEY_LEFT :  stateMachine->setArrowState(LEFT , true);
                                        twi->faceRight(false);
                                        twi->setIntendedXAcc(-VELOCITY_SCALE/4);
                                        break;
        case GLUT_KEY_UP :    stateMachine->setArrowState(UP   , true); break;
        case GLUT_KEY_DOWN :  stateMachine->setArrowState(DOWN , true); break;
        default: break;
    }
}

void Game::specialUp(int key, int x, int y)
{
    switch(key) {
        case GLUT_KEY_RIGHT : stateMachine->setArrowState(RIGHT, false);
                                        if(stateMachine->getArrowStates(LEFT)) {
                                            twi->faceRight(false);
                                            twi->setIntendedXAcc(-VELOCITY_SCALE/4);
                                        }
                                        else
                                            twi->setIntendedXAcc(0);
                                        break;
        case GLUT_KEY_LEFT :  stateMachine->setArrowState(LEFT , false);
                                        if(stateMachine->getArrowStates(RIGHT)) {
                                            twi->faceRight(true);
                                            twi->setIntendedXAcc(VELOCITY_SCALE/4);
                                        }
                                        else
                                            twi->setIntendedXAcc(0);
                                        break;
        case GLUT_KEY_UP :    stateMachine->setArrowState(UP   , false); break;
        case GLUT_KEY_DOWN :  stateMachine->setArrowState(DOWN , false); break;
        default: break;
    }
}

void Game::reshape(int width, int height)
{
    Rect bg_bounds = level->getBgSize();
    offsetX = artist->windowWidth()/2 - bg_bounds.w /2;
    offsetY = artist->windowHeight()/2 - bg_bounds.h /2;
}

void Game::assumeControl(StateLoopIFC*& currentActor)
{
    cout << "--- GAME IS ASSUMING CONRTOL OF GLUT MAIN LOOP ---\n";
    currentActor = this;
    reshape(artist->windowWidth(), artist->windowHeight());
}
